# imports
import os
import fnmatch

from ibis_birdbrain.tools import tool


# helpers
def read_gitignore(gitignore_path):
    with open(gitignore_path, "r") as f:
        lines = f.readlines()
    return [line.strip() for line in lines if not line.startswith("#")]


def is_ignored(path, ignore_patterns):
    for pattern in ignore_patterns:
        if fnmatch.fnmatch(path, pattern):
            return True
    return False


# tools
@tool
def read_file(path: str) -> str:
    """Reads a file and returns its content."""
    # TODO: fix hack
    if path.startswith("https://") or path.startswith("http://"):
        from ibis_birdbrain.tools.internet import webpage_to_str

        return webpage_to_str(path)
    path = os.path.expanduser(path)
    with open(path, "r") as f:
        return f.read()


@tool
def list_files_and_dirs(
    path: str = ".", depth: int = -1, additional_ignore_dirs: list = []
):
    """Lists all files and directories in a directory."""
    path = os.path.expanduser(path)
    files_dirs_list = []
    home = os.path.expanduser("~")
    gitignore_path = os.path.join(home, ".gitignore")

    if os.path.exists(gitignore_path):
        gitignore_patterns = read_gitignore(gitignore_path)
    else:
        gitignore_patterns = []

    ignore_dirs = [".git"] + additional_ignore_dirs

    for root, dirs, files in os.walk(path):
        if root.count(os.sep) >= depth and depth != -1:
            dirs.clear()  # Clear directories list to prevent further depth traversal.

        dirs[:] = [
            d
            for d in dirs
            if not is_ignored(d, ignore_dirs) and not is_ignored(d, gitignore_patterns)
        ]

        for file in files:
            file_path = os.path.join(root, file)
            if not is_ignored(file_path, gitignore_patterns):
                files_dirs_list.append(file_path)

        for d in dirs:
            dir_path = os.path.join(root, d)
            if not is_ignored(dir_path, gitignore_patterns):
                files_dirs_list.append(dir_path)

    return files_dirs_list


@tool
def write_file(path: str, content: str, overwrite: bool = False) -> str:
    """Writes a file."""
    path = os.path.expanduser(path)
    mode = "w" if overwrite else "a"
    with open(path, mode) as f:
        if mode == "a":
            f.write("\n\n")
        f.write(content)

    return f"File written (mode: {mode}) to {path}"



# imports
from typing import Callable

from ibis_birdbrain.tools import tool
from ibis_birdbrain.tools.code import text_to_python, run_python_code


# tools
@tool
def use_github_cli(text: str = "open the current repo") -> str:
    """Uses the GitHub CLI to fulfill the request via text."""
    text = f"""
    Write a simple Python function to run the `os.system` command using the
    GitHub CLI to fulfill the request via text, e.g.:

    ```python
    import subprocess

    out = subprocess.run(["gh", "repo", "view", "--web"], capture_output=True)
    print(out.stdout.decode())
    ```

    for:

    {text}
    """
    code = text_to_python(text)
    res = run_python_code(code)
    return f"Ran the GitHub CLI successfully! Result: {res}"



# imports
import ibis

# create connection to duckdb
con = ibis.connect("duckdb://imdb.ddb")

# fetch data from ibis examples
imdb_name_basics = ibis.examples.imdb_name_basics.fetch()
imdb_title_akas = ibis.examples.imdb_title_akas.fetch()
imdb_title_basics = ibis.examples.imdb_title_basics.fetch()
imdb_title_crew = ibis.examples.imdb_title_crew.fetch()
imdb_title_episode = ibis.examples.imdb_title_episode.fetch()
imdb_title_principals = ibis.examples.imdb_title_principals.fetch()
imdb_title_ratings = ibis.examples.imdb_title_ratings.fetch()

# create tables in con
con.create_table("imdb_name_basics", imdb_name_basics.to_pyarrow(), overwrite=True)
con.create_table("imdb_title_akas", imdb_title_akas.to_pyarrow(), overwrite=True)
con.create_table("imdb_title_basics", imdb_title_basics.to_pyarrow(), overwrite=True)
con.create_table("imdb_title_crew", imdb_title_crew.to_pyarrow(), overwrite=True)
con.create_table("imdb_title_episode", imdb_title_episode.to_pyarrow(), overwrite=True)
con.create_table(
    "imdb_title_principals", imdb_title_principals.to_pyarrow(), overwrite=True
)
con.create_table("imdb_title_ratings", imdb_title_ratings.to_pyarrow(), overwrite=True)



# imports
from ibis_birdbrain.tools import tool

from io import StringIO
from contextlib import redirect_stdout, redirect_stderr

from ibis_birdbrain.functions import (
    gen_python_code,
    fix_python_code,
)


# tools
@tool
def text_to_python(question: str) -> str:
    """Returns a Python code given text."""
    return gen_python_code(question)


@tool
def fix_python_error(code: str, error: str) -> str:
    """Fixes a Python error in the code."""
    return fix_python_code(code, error)


@tool
def python_function_to_udf(code: str) -> str:
    """Converts a Python function to an Ibis UDF."""
    return f"""
import ibis

@ibis.udf.scalar.python
{code}""".strip()


@tool
def run_python_code(code: str) -> str:
    """Execute Python code as a string and return the output"""
    code = f"""
import subprocess

def run_code():
    {code}

out = subprocess.run(run_code(), capture_output=True, text=True)

return out.stdout, out.stderr"""

    try:
        local_vars = {}
        exec(code, globals(), local_vars)
        stdout, stderr = local_vars["out"]
    except Exception as e:
        stdout = ""
        stderr = f"Error: {e}"

    return f"Code ran successfully!\nstdout: {stdout}\nstderr: {stderr}"



# imports
from typing import Callable

from ibis_birdbrain.tools import tool
from ibis_birdbrain.tools.internet import webpage_to_str
from ibis_birdbrain.tools.text import summarize_text, Summary
from ibis_birdbrain.tools.filesystem import read_file, write_file


# tools
@tool
def read_webpage_and_summarize(url: str = "https://ibis-project.org") -> Summary:
    """Read a webpage and summarize it."""
    return summarize_text(webpage_to_str(url))


@tool
def read_file_and_summarize(
    path: str,
    file_write: bool = True,
    filename: str = "notes.md",
    file_overwrite: bool = False,
) -> Summary:
    """Read a file and summarize it."""
    summary = summarize_text(read_file(path))
    if file_write:
        return write_file(filename, str(summary), overwrite=file_overwrite)

    return summary


@tool
def read_files_and_summarize(
    files: list[str],
    file_write: bool = True,
    filename: str = "notes.md",
    file_overwrite: bool = False,
) -> str:
    """Read a list of files and summarize them."""
    for file in files:
        summary = summarize_text(read_file(file))
        if summary is None:
            continue
        if file_write:
            write_file(filename, f"filename: {file}")
            write_file(filename, str(summary), overwrite=file_overwrite)

    return f"Files {files} summarized and written to {filename}."


@tool
def run_on_each_file(
    files: list[str],
    func: Callable,
) -> str:
    """Run a function on each file in a list of files."""
    return "\n".join([func(file) for file in files])



# imports
from enum import Enum
from marvin import ai_classifier


# classifiers
@ai_classifier
class EvalPythonCode(Enum):
    """Classifies Python code as valid or invalid."""

    INVALID = False
    VALID = True


@ai_classifier
class EvalSQLCode(Enum):
    """Classifies SQL code as valid or invalid."""

    INVALID = False
    VALID = True


@ai_classifier
class EvalTextForQuery(Enum):
    """Classifies text as requiring a SQL query to answer or not."""

    NOT_QUERY = False
    QUERY = True



# imports
import os
import toml
import typer
import marvin

from dotenv import load_dotenv
from typing_extensions import Annotated

from ibis_birdbrain.commands.ai import ai_run
from ibis_birdbrain.commands.testing import testing_run

# load .env file
load_dotenv(os.path.expanduser("~/.ibis_birdbrain/.env"))
marvin.settings.llm_model = "azure_openai/gpt-4-32k"

# load config
try:
    config = toml.load(os.path.expanduser("~/.ibis_birdbrain/config.toml"))
except FileNotFoundError:
    config = {}

# typer config
app = typer.Typer(no_args_is_help=True)

# global state
state = {"config": config}


# global options
def version(value: bool):
    if value:
        version = toml.load("pyproject.toml")["project"]["version"]
        typer.echo(f"{version}")
        raise typer.Exit()


# subcommands
@app.command()
def test():
    """
    test
    """
    testing_run()


@app.command()
def ai(
    interactive: Annotated[
        bool, typer.Option("--interactive", "-i", help="run ai in interactivce mode")
    ] = False,
    learn_spanish: Annotated[
        bool, typer.Option("--learn-spanish", "-l", help="run in learn Spanish mode")
    ] = False,
):
    """
    ai
    """
    ai_run(state=state, interactive=interactive, learn_spanish=learn_spanish)


# main
@app.callback()
def cli(
    version: bool = typer.Option(
        None, "--version", help="Show version.", callback=version, is_eager=True
    ),
):
    version = version
    # Do other global stuff, handle other global options here
    return


## main
if __name__ == "__main__":
    typer.run(cli)



# imports
from marvin.prompts.library import System, User, ChainOfThought


# systems
class ibis_birdbrainSystem(System):
    content: str = """A bot named ibis_birdbrain that serves as a digital assitant for exploratory data analysis, data engineering, and data science with Ibis and AI.

Notice your tools -- you can access a data platform via Ibis, run SQL, generate and run Python code, search the internet, summarize things, and more!

You are to use your tools to assist the user, never fabricate information, always query data for answers, and guide the user based on examples in the filesystem."""


class UserPreferencesSystem(System):
    content: str = """Use simple, plain language.

YOU MUST never captialize `ibis_birdbrain` in any form, ever.

YOU MUST only capitalize the first word in a heading. Use markdown format.
"""


class FixesSystem(System):
    content: str = """YOU MUST use your tools and MUST NOT answer questions about data without querying data.

YOU MUST only open a URL in the browser once per ask."""


class CiteSourceSystem(System):
    content: str = """YOU MUST cite your sources
for any information you provide. You can use the following format:

    [1] https://www.example.com/a-first-page

    [2] https://www.example.com/another-page

In most cases, you should cite AT LEAST three sources."""


class LearnSpanishSystem(System):
    content: str = """For this conversation, you are to always respond in Spanish despite the user's language to help them learn Spanish. Disregard if explicitly told."""



# imports
import requests

from itertools import islice
from html2text import html2text

from ibis_birdbrain.tools import tool


# tools
@tool
def search_internet(
    query: str = "what is the Ibis project?", n_results: int = 8
) -> list[dict[str, str | None]]:
    """Searches the internet for the given query."""
    from duckduckgo_search import DDGS

    ddgs = DDGS()
    return [r for r in islice(ddgs.text(query, backend="lite"), n_results)]


@tool
def webpage_to_str(url: str = "https://ibis-project.org") -> str:
    """Reads a webpage link into a string. Useful for summarizing webpages."""
    response = requests.get(url)
    return html2text(response.text)


@tool
def open_browser(url: str) -> str:
    """Opens the URL in a web browser."""
    try:
        import webbrowser

        webbrowser.open(url.strip("/"))
    except Exception as e:
        return str(e)

    return f"Opened {url} in the user's web browser."


@tool
def let_me_google_that_for_you(query: str) -> None:
    """Snarky tool that opens a web browser with a LMGTFY link."""
    open_browser("https://letmegooglethat.com/?q=" + query)



# imports
from marvin.tools import tool

from ibis_birdbrain.tools.internet import (
    open_browser,
    search_internet,
    webpage_to_str,
)
from ibis_birdbrain.tools.text import summarize_text, translate_text
from ibis_birdbrain.tools.code import text_to_python, fix_python_error, run_python_code
from ibis_birdbrain.tools.filesystem import read_file, list_files_and_dirs, write_file
from ibis_birdbrain.tools.ibis_birdbrain import list_tables, query_table, get_table_schema
from ibis_birdbrain.tools.advanced import (
    read_files_and_summarize,
    read_webpage_and_summarize,
)
from ibis_birdbrain.tools.github import use_github_cli


# tools
tools = [
    # internet
    open_browser,
    search_internet,
    webpage_to_str,
    # text
    summarize_text,
    translate_text,
    # filesystem
    read_file,
    list_files_and_dirs,
    write_file,
    # code
    text_to_python,
    fix_python_error,
    run_python_code,
    # data
    list_tables,
    query_table,
    get_table_schema,
    # advanced
    read_files_and_summarize,
    read_webpage_and_summarize,
    # github
    # use_github_cli,
]

__all__ = ["tool", "tools"]



# imports
from marvin import ai_model
from pydantic import BaseModel, Field


# models
@ai_model
class Translate(BaseModel):
    """A translation of input text from one language to another"""

    from_: str = Field(description="The language to translate from")
    to: str = Field(description="The language to translate to")
    original: str = Field(description="The original text")
    translated: str = Field(description="The translated text")


@ai_model
class SQL(BaseModel):
    """A SQL query"""

    original: str = Field(description="The original text")
    user_intent: str = Field(description="The user's intent with the text")
    sql: str = Field(description="The SQL SELECT query to run based on the text")


@ai_model
class Summary(BaseModel):
    """A summary of input text"""

    date: str = Field(description="The date of the text")
    author: str = Field(description="The author of the text")
    description: str = Field(description="The description of the text")
    summary: str = Field(description="The one to two paragraph summary of the text")
    key_points: list[str] = Field(description="The key points of the text, if any")
    sentiment: str = Field(description="The sentiment of the text")
    sentiment_score: float = Field(description="The sentiment score of the text")
    biases: list[str] = Field(
        description="The obvious biases of the text's author, if any"
    )

    def __str__(self):
        return f"""Summary(
    date: {self.date},
    author: {self.author},
    description: {self.description},
    summary: {self.summary},
    key_points: {self.key_points},
    sentinment: {self.sentiment},
    sentinment_score:{self.sentiment_score},
    biases:{self.biases}\n)"""



# imports
from ibis_birdbrain import AI

from ibis_birdbrain.tools import tools
from ibis_birdbrain.states import ibis_birdbrainState
from ibis_birdbrain.systems import (
    ibis_birdbrainSystem,
    FixesSystem,
    CiteSourceSystem,
    UserPreferencesSystem,
)

# variables
state = ibis_birdbrainState()
prompts = [FixesSystem(), CiteSourceSystem(), UserPreferencesSystem()]
description = ibis_birdbrainSystem().content


# create bot
bot = AI(
    name=state.name, description=description, tools=tools, prompts=prompts, state=state
)



# imports
from pydantic import BaseModel


# states
class ibis_birdbrainState(BaseModel):
    """State of ibis_birdbrain"""

    # info about bot
    name: str = "ibis_birdbrain"
    creator: str = "Ibis developers"
    version: str = "infinity"

    # links to open
    ibis_docs: str = "https://www.ibis-project.org"
    ibis_github: str = "https://github.com/ibis-project/ibis"
    marvin_docs: str = "https://www.askmarvin.ai/components/overview"
    marvin_github: str = "https://github.com/prefectHQ/marvin"
    self_source_code: str = "https://github.com/ibis-project/ibis-ibis_birdbrain"

    # additional links
    additional_links: list[str] = []

    # research info
    analysis_topic: str = "data"
    analysis_description: str = ""
    analysis_abstract: str = ""
    analysis_references: list[str] = []
    analysis_analysis: str = ""



def ai_run(state={}, interactive=False, learn_spanish=False):
    from ibis_birdbrain import Console
    from ibis_birdbrain.ai import bot
    from ibis_birdbrain.systems import LearnSpanishSystem

    if learn_spanish:
        bot.ai.additional_prompts.append(LearnSpanishSystem())

    console = Console()
    console.print(f"ibis_birdbrain: ", style="bold violet blink")
    console.print(f"access to: {bot.name}")
    console.print(f"state: {state}")

    if interactive:
        import ibis
        import IPython

        from rich import print

        from ibis_birdbrain.tools.ibis_birdbrain import con, tables

        ibis.options.interactive = True

        IPython.embed(colors="neutral")



# imports
import os
import marvin

from dotenv import load_dotenv
from marvin import AIApplication as Bot
from rich.console import Console

# load .env
load_dotenv(os.path.expanduser("~/.birdbrain/.env"))

# hacks
# TODO: do this elsewhere?
model = "azure_openai/gpt-4-32k"
marvin.settings.llm_model = model

# TODO: hack? check if bug in Marvin
marvin.settings.azure_openai.api_key = os.getenv("MARVIN_AZURE_OPENAI_API_KEY")
marvin.settings.azure_openai.api_base = os.getenv("MARVIN_AZURE_OPENAI_API_BASE")
marvin.settings.azure_openai.deployment_name = os.getenv(
    "MARVIN_AZURE_OPENAI_DEPLOYMENT_NAME"
)
marvin.settings.azure_openai.api_type = "azure"


# classes
class AI:
    def __init__(self, name, description, tools, prompts, state):
        ai = Bot(
            name=name,
            description=description,
            tools=tools,
            additional_prompts=prompts,
            state=state,
        )

        self.ai = ai
        self.name = name
        self.tools = tools
        self.prompts = prompts
        self.description = description

        self.console = Console()

    def __call__(self, text) -> str | None:
        res = self.ai(text).content
        if self.name == "ibis_birdbrain":
            self.console.print(f"ibis_birdbrain:\n\n", style="bold violet blink", end="")
            self.console.print(res)
        else:
            return res

    def __repr__(self):
        return f"<AI {self.name}>"

    def read_excel(self, filename):
        pass

    def sql(self, text):
        import ibis
        from ibis_birdbrain.functions import choose_table_name, gen_sql_query, fix_sql_query

        con = ibis.connect("duckdb://imdb.ddb")

        tables = list(set(con.list_tables()))
        table_name = choose_table_name(text, options=tables)
        table_schema = str(con.table(table_name).schema())

        if table_name not in tables:
            raise ValueError(f"Table {table_name} not found in {tables}")
        sql = gen_sql_query(table_name, table_schema, text).strip(";")
        res = None
        try:
            res = con.table(table_name, database="imdb").sql(sql)
        except Exception as e:
            sql = fix_sql_query(table_name, table_schema, sql, str(e)).strip(";")
            try:
                res = con.table(table_name, database="imdb").sql(res)
            except Exception as e:
                raise ValueError(f"Could not execute SQL: {res} with error: {e}")

        self.console.print(f"ran SQL:\n{sql}", style="bold violet")
        return res


__all__ = ["AI", "Bot", "Console"]



# imports
from marvin import ai_fn


# functions
@ai_fn
def choose_table_name(text: str, options: list[str]) -> str:
    """Given text, returns the table name to query."""
    text, options = text, options
    return ""


@ai_fn
def gen_sql_query(
    table_name: str, table_schema: str, question: str = "how many rows are there?"
) -> str:
    """Given a question about a table in a database, returns the SQL query
    string. MUST ONLY use SELECT and a single table of the input tablename."""
    table_name, table_schema, question = table_name, table_schema, question
    return ""


@ai_fn
def fix_sql_query(table_name: str, table_schema: str, query: str, error: str) -> str:
    """Given a SQL query and an error, returns the fixed SQL query."""
    table_name, table_schema, query, error = table_name, table_schema, query, error
    return ""


@ai_fn
def gen_python_code(description: str) -> str:
    """Generates a Python code for the input description. Keep it simple,
    type-annotated, and using minimal external dependencies. It may be fine to
    install a few dependencies.

    If a function is generated, ensure the function is called.
    """
    description = description
    return ""


@ai_fn
def fix_python_code(code: str, error: str) -> str:
    """Given a Python code and an error, returns the fixed Python
    function."""
    code, error = code, error
    return ""



# imports
from ibis_birdbrain.models import Summary, Translate

from ibis_birdbrain.tools import tool


# tools
@tool
def summarize_text(text: str) -> Summary:
    """Summarizes text"""
    return Summary(text)


@tool
def translate_text(
    text: str, from_: str = "English (US)", to: str = "Spanish (Mexico)"
) -> Translate:
    """Translates text"""
    text = f"translate `{text}` from {from_} to {to}"
    return Translate(text)



# imports
import ibis

from ibis_birdbrain.tools import tool
from ibis_birdbrain.functions import choose_table_name, gen_sql_query, fix_sql_query

# setup Ibis
con = ibis.connect("duckdb://imdb.ddb")
tables = list(set(con.list_tables()))


# tools
@tool
def get_table_schema(table_name: str) -> str:
    """Returns the schema of a table"""
    if table_name not in tables:
        raise ValueError(f"Table {table_name} not found in {tables}")
    return str(con.table(table_name).schema())


@tool
def list_tables() -> list[str]:
    """Returns a list of available tables to query"""
    return list(set(con.list_tables()))


@tool
def query_table(question: str) -> str:
    """Queries the table in the database to answer the question"""
    table_name = choose_table_name(question, options=tables)
    table_schema = get_table_schema(table_name)
    if table_name not in tables:
        raise ValueError(f"Table {table_name} not found in {tables}")
    sql = gen_sql_query(table_name, table_schema, question).strip(";")
    try:
        res = con.table(table_name).sql(sql)
    except Exception as e:
        sql = fix_sql_query(table_name, table_schema, sql, str(e)).strip(";")
        try:
            res = con.table(table_name).sql(res)
        except Exception as e:
            raise ValueError(f"Could not execute SQL: {res} with error: {e}")

    return str(res)
